/**
  *  \file client/tiles/fleetmemtile.cc
  *  \brief Fleet Member Tile
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/fleetmemtile.h"
#include "client/widgets/fleetmemlist.h"
#include "game/ship.h"
#include "game/global.h"
#include "ui/widgets/button.h"
#include "util/nls.h"
#include "ui/util.h"
#include "game/objsel.h"


enum { NLINES = 10 };

WFleetMemberTile::WFleetMemberTile(GObjectSelection& selection)
    : WTile(GfxPoint(315, NLINES*font_heights[FONT_NORMAL] + 20 + 5)),
      selection(selection),
      list(0)
{
    init();
}

WFleetMemberTile::~WFleetMemberTile()
{ }

void
WFleetMemberTile::init()
{
    // Make sure we can receive focus
    setState(st_Disabled, false);

    // Default layout
    setLayout(LeftJustified, 5, true);

    // Create list
    list = &holder.add(new WFleetMemberList(NLINES, 305));
    addControl(GfxRect(0, 20+4, 305, font_heights[FONT_NORMAL]*NLINES), *list);
    list->sig_moved.add(this, &WFleetMemberTile::onMove);
    list->setFocusedWidget();

    // Buttons
    addControl(GfxRect(0, 0, 35, 20), holder.add(new UIButton(_("Del"), SDLK_DELETE)));
    addControl(GfxRect(40, 0, 20, 20), holder.add(new UIButton("B", 'b')));
    addControl(GfxRect(65, 0, 20, 20), holder.add(new UIButton("T", 't')));
    addControl(GfxRect(90, 0, 20, 20), holder.add(new UIButton("P", 'p'))); // was 'S' in PCC 1.x

    addControl(GfxRect(285, 0, 20, 20), holder.add(new UIButton("F", 'f')));

    // Handle fleet change
    conn_sel_pos_changed = selection.sig_current_changed.add(this, &WFleetMemberTile::onFleetChanged);
    onFleetChanged();

    // Handle turn change
    conn_turn_changed = sig_turn_changed.add(this, &WFleetMemberTile::onTurnChanged);
    onTurnChanged();
}

void
WFleetMemberTile::onMove()
{
    // Cursor position changed on fleet list. Point global ship selection to the member.
    int sid = list->getCurrentMemberId();
    GUniverse* univ = selection.getCurrentUniverse();
    if (sid != 0
        && univ != 0
        && univ->ty_played_ships.isValidIndex(sid)
        && current_ship_type.getObjectType() == &univ->ty_played_ships)
    {
        current_ship_selection.setCurrentIndex(sid);
    }
}

void
WFleetMemberTile::onFleetChanged()
{
    // A new object has been selected on the underlying object type.
    // If it is a fleet, load it.
    int fid = 0;
    GShip* sh = dynamic_cast<GShip*>(selection.getCurrentObject());
    if (sh != 0 && sh->getId() == sh->getFleetNumber()) {
        fid = sh->getId();
    }

    // setFleet will call onMove()
    list->setFleet(fid);
}

/** Called when the displayed turn changed. This means we have to
    re-hook the universe change hook. */
void
WFleetMemberTile::onTurnChanged()
{
    conn_univ_pre_update = getDisplayedTurn().getCurrentUniverse().sig_pre_update.add(this, &WFleetMemberTile::onUniverseChanged);
    onUniverseChanged();
}

/** Called when the universe changed. */
void
WFleetMemberTile::onUniverseChanged()
{
    /* Check the list of displayed members, not the list of actual members, to catch
       ships that leave the fleet. If a ship joins the fleet, the leader will be marked
       dirty, so we catch them this way. */
    if (list->isAnyMemberDirty()) {
        list->setFleet(list->getFleet());
        list->drawWidget();
    }
}


void
WFleetMemberTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Static);
    ctx.setTextAlign(2, 1);
    outText(ctx, 280, 10, _("FCode"));

    GfxRect r(list->getExtent());
    r.moveByRev(getExtent().x, getExtent().y);
    r.grow(1, 1);
    drawFrameDown(can, r);
}
