/**
  *  \file client/tiles/vishull.cc
  *  \brief Visual Scanner: Hull Information Tile
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/vishull.h"
#include "cpluslib/format.h"
#include "game/ship.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "util/unicodechars.h"
#include "util/misc.h"

/************************ WVisualScannHullInfoTile ***********************/

WVisualScanHullInfoTile::WVisualScanHullInfoTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(font_sizes[FONT_NORMAL].scaledBy(10, 6), selection)
{ }

void
WVisualScanHullInfoTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();
    ctx.setColor(UIColor::tc_Static);

    GShip* s;
    mp16_t hid;
    if ((s = dynamic_cast<GShip*>(getCurrentObject())) != 0 && (hid = s->getHullId()).isKnown()) {
        const GHull& h = getHull(hid);
        const int w = extent.w;
        const int hh = font_heights[FONT_NORMAL];

        // Line 1: "Hull mass: nn kt"
        string_t text = format(_("Hull mass: %d kt"), numToString(h.getMass()));
        outTextF(ctx, 0, 0, w, text);

        // Line 2:
        //   Cargo: a/b kt
        //   Max cargo: x kt
        if (s->getShipKind() == GShip::CurrentShip)
            text = format(_("Cargo: %d/%d kt"), numToString(h.getCargo() - s->getFreeCargo()), numToString(h.getCargo()));
        else
            text = format(_("Max Cargo: %d kt"), numToString(h.getCargo()));
        outTextF(ctx, 0, hh, w, text);

        // Line 3:
        //   Fuel: a/b kt
        //   Max fuel: x kt
        if (s->getShipKind() == GShip::CurrentShip)
            text = format(_("Fuel: %d/%d kt"), numToString(s->getCargo(el_Neutronium)), numToString(h.getFuel()));
        else
            text = format(_("Max Fuel: %d kt"), numToString(h.getFuel()));
        outTextF(ctx, 0, 2*hh, w, text);

        // Line 4:
        //   3xBeamTyp
        //   Max Beams: x
        //   No beams
        // FIXME: handle history
        if (s->getShipKind() == GShip::CurrentShip) {
            if (s->getNumBeams() > 0)
                text = format("%d" UTF_TIMES "%s", s->getNumBeams(), getBeam(s->getBeamType()).getName());
            else
                text = _("No beams");
        } else {
            if (h.getMaxBeams() > 0)
                text = format(_("Max Beams: %d"), h.getMaxBeams());
            else
                text = _("No beams");
        }
        outTextF(ctx, 0, 3*hh, w, text);

        // Line 5:
        //   2xTorpTyp
        //   Max torps: x
        //   No torps
        //   Fighter bays: x
        //   No fighter bays
        if (s->getShipKind() == GShip::CurrentShip) {
            if (s->getNumTorpLaunchers() > 0) {
                text = format("%d" UTF_TIMES "%s", s->getNumTorpLaunchers(), getTorp(s->getTorpType()).getName());
            } else if (s->getNumBays() > 0) {
                text = format(_("Fighter bays: %d"), s->getNumBays());
            } else if (h.getNumBays() > 0) {
                text = _("No fighter bays");
            } else if (h.getMaxLaunchers() > 0) {
                text = _("No torps");
            } else {
                text.clear();
            }
        } else {
            if (h.getNumBays() > 0)
                text = format(_("Fighter bays: %d"), h.getNumBays());
            else if (h.getMaxLaunchers() > 0)
                text = format(_("Max torps: %d"), h.getMaxLaunchers());
            else
                text.clear();
        }
        outTextF(ctx, 0, 4*hh, w, text);

        // Line 6:
        //   FCode: foo
        if (s->getShipKind() == GShip::CurrentShip)
            text = format(_("FCode: %s"), s->getFCode());
        else
            text.clear();
        outTextF(ctx, 0, 5*hh, w, text);
    } else {
        // Hull not known: completely blank
        drawBackground(ctx, getInnerSize(0, 0));
    }
}

void
WVisualScanHullInfoTile::getLayoutInfo(LayoutInfo& info)
{
    WObjectSelectionObserverTile::getLayoutInfo(info);
    info.flags |= li_FlexibleH;
}
