/**
  *  \file client/widgets/consoleview.cc
  *  \brief Console View
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/consoleview.h"
#include "ui/uiglobal.h"
#include "util/console.h"
#include "gfx/palette.h"
#include "cpluslib/format.h"

/****************************** WConsoleView *****************************/

/** Constructor.
    \param console Console to display
    \param lines Lines to display (window height) */
WConsoleView::WConsoleView(Console& console, int lines)
    : console(console),
      conn_console(console.console_changed.add(this, &WConsoleView::onUpdate)),
      display_lines(lines),
      scrollback_lines(0)
{ }

/** Destructor. */
WConsoleView::~WConsoleView()
{ }

/** Handle console update. Called from the console's console_changed signal. */
void
WConsoleView::onUpdate(const Console& /*console*/, int added_lines)
{
    if (added_lines == 0) {
        /* console cleared, scroll to end */
        scrollback_lines = 0;
        drawWidget(synchronous);
    } else {
        /* some lines added to the console */
        if (scrollback_lines != 0)
            scrollback_lines += added_lines;
        if (scrollback_lines > 0)
            limitScrollback();
        drawWidget(synchronous);
    }
}

/** Limit scrollback_lines parameter to prevent scrolling too far. */
void
WConsoleView::limitScrollback()
{
    int lines_displayed = getExtent().h / font_heights[FONT_NORMAL];
    int lines_in_buffer = console.getScrollbackFill();

    /* Limit scrollback */
    int limit = lines_in_buffer - lines_displayed;
    if (scrollback_lines > limit)
        scrollback_lines = limit;

    /* Cannot scroll into future */
    if (scrollback_lines < 0)
        scrollback_lines = 0;
}

void
WConsoleView::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    int lines_displayed = getExtent().h / font_heights[FONT_NORMAL];
    for (int i = 0; i < lines_displayed; ++i) {
        /* Clear line */
        drawBackground(ctx, GfxRect(getExtent().x, getExtent().y + font_heights[FONT_NORMAL]*i,
                                    getExtent().w, font_heights[FONT_NORMAL]));

        /* Line already written? */
        int index = scrollback_lines + (lines_displayed-1 - i);
        if (index < console.getScrollbackFill()) {
            /* Fetch line from console */
            const Console::Message& m = console.getLine(index);

            /* Find message color */
            int color = (m.format & ConsoleColorMask);
            if (color != 0 && color < UIColor::COUNT) {
                /* Message specifies a color */
                ctx.setColor(UIColor::Color(color));
            } else if (m.mc == LOG_DEBUG) {
                /* Dim debug messages */
                ctx.setColor(UIColor::tc_Faded);
            } else if (m.mc == LOG_ERROR) {
                /* Highlight error messages */
                ctx.setColor(UIColor::tc_Red);
            } else {
                /* Warning and Info (i.e. normal) messages */
                ctx.setColor(UIColor::tc_Static);
            }

            /* Choose font */
            if ((m.format & (ConsoleRecallable | ConsoleRightJustified)) == ConsoleRecallable) {
                /* Input */
                ctx.useFont(*fonts[FONT_NORMAL_BOLD]);
            } else {
                /* Everything else */
                ctx.useFont(*fonts[FONT_NORMAL]);
            }

            /* Display message with correct alignment */
            const int y = getExtent().y + font_heights[FONT_NORMAL]*i;
            const int w = getExtent().w;
            if (m.format & ConsoleRightJustified) {
                /* Right-justified */
                ctx.setTextAlign(2, 0);
                outTextF(ctx, getExtent().getRightX(), y, w, m.text);
            } else if (m.format & ConsoleCentered) {
                /* Centered: if it has a tab, center at that; otherwise, center as a whole */
                string_t::size_type n = m.text.find('\t');
                int x = getExtent().getCenter().x;
                if (n != string_t::npos) {
                    ctx.setTextAlign(2, 0);
                    outTextF(ctx, x, y, w/2, m.text.substr(0, n));
                    ctx.setTextAlign(0, 0);
                    outTextF(ctx, x, y, w/2, m.text.substr(n+1));
                } else {
                    ctx.setTextAlign(1, 0);
                    outTextF(ctx, x, y, w, m.text);
                }
            } else {
                /* Default: left-justified */
                ctx.setTextAlign(0, 0);
                outTextF(ctx, getExtent().x, y, w, m.text);
            }
        }
    }

    /* Show scrollback indicator */
    if (scrollback_lines > 0) {
        string_t txt = format("[-%d]", scrollback_lines);
        int wi = ctx.getFont().getTextWidth(txt);
        int rx = getExtent().getRightX() - wi;

        drawSolidBar(ctx, GfxRect(rx, getExtent().y, wi, font_heights[FONT_NORMAL]), UIColor::tc_Red);
        ctx.setColor(UIColor::tc_White);
        ctx.setTextAlign(0, 0);
        outText(ctx, rx, getExtent().y, txt);
    }

    can.update(getExtent());
}

void
WConsoleView::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = font_sizes[FONT_NORMAL].scaledBy(30, display_lines);
    info.flags = li_FlexibleH;
}

/** Scroll to end. */
void
WConsoleView::scrollToEnd()
{
    if (scrollback_lines != 0) {
        scrollback_lines = 0;
        drawWidget();
    }
}

/** Scroll after end, such that console appears empty. */
void
WConsoleView::scrollAfterEnd()
{
    if (scrollback_lines != -display_lines) {
        scrollback_lines = -display_lines;
        drawWidget();
    }
}

/** Scroll backward in time (=up).
    \param lines Number of lines to scroll backward. Can be negative to scroll forward */
void
WConsoleView::scrollBack(int lines)
{
    int old = scrollback_lines;
    scrollback_lines += lines;
    limitScrollback();
    if (old != scrollback_lines)
        drawWidget();
}

/** Set synchronous mode. If synchronous mode is enabled, changes to the console are
    redrawn immediately, not in the next update cycle.
    \param mode New status */
void
WConsoleView::setSynchronousMode(bool mode)
{
    if (mode != synchronous) {
        synchronous = mode;
        if (mode && hasState(st_WantRedraw))
            drawWidget(true);
    }
}

/************************ WConsoleViewWithHandler ************************/

WConsoleViewWithHandler::WConsoleViewWithHandler(Console& console, int lines)
    : WConsoleView(console, lines)
{ }

bool
WConsoleViewWithHandler::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        if (e.key.code == SDLK_UP) {
            setSelectedWidget();
            scrollBack(1);
            return true;
        }
        if (e.key.code == SDLK_DOWN) {
            setSelectedWidget();
            scrollBack(-1);
            return true;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

