/**
  *  \file game/greg.cc
  *  \brief Game Registries
  *
  *  This module is called "Game Registries" so it can also be a home
  *  to a possible future gamestat.dat parser.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/greg.h"
#include "game/gamedirs.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/random.h"
#include "cpluslib/misc.h"

namespace {
    enum { Num_Slots = 40 };
    struct Data {
        char slot[2];
        char timestamps[Num_Slots][18];
        char templock[Num_Slots][44];
    };
    CTASSERT(sizeof(Data) == 2 + 18*Num_Slots + 44*Num_Slots);
}

/** Update Winplan game registry file (snooker.dat). Winplan keeps a
    registry of the 40 most recently seen timestamps, and refuses to
    open games not in that list. This function adds a timestamp to the
    list if required.

    PCC currently never reads these values. */
void
updateGameRegistry(GTimestamp ts)
{
    Data file_data;

    // Nothing to do if no winplan directory
    if (!winplan_dir)
        return;

    // Only update file if it exists
    Ptr<Stream> s = winplan_dir->openFileNT("snooker.dat", Stream::C_READ | Stream::C_WRITE);
    if (!s)
        return;

    // If the file is truncated, we'll happily repair it
    zeroFill(file_data);
    s->read(file_data.slot, sizeof(file_data));

    // Look for our entry
    for (int i = 0; i < Num_Slots; ++i)
        if (ts == file_data.timestamps[i])
            return;

    // Entry not found, so add it
    uint16 slot = getInt16(file_data.slot);
    ++slot;
    if (slot < 1 || slot > Num_Slots)
        slot = 1;
    ts.storeRawData(file_data.timestamps[slot-1]);
    for (int byte = 0; byte < 44; ++byte)
        file_data.templock[slot-1][byte] = myRandom(256);
    storeInt16(file_data.slot, slot);

    // Write back file
    s->seek(0);
    s->writeT(file_data.slot, sizeof(file_data));
}
