/**
  *  \file game/movement.cc
  *  \brief Whole-Universe Movement Prediction
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include "game/movement.h"
#include "game/univ.h"
#include "game/mission.h"
#include "game/shippredict.h"
#include "game/cost.h"
#include "cpluslib/assert.h"

GMovementPredictor::GMovementPredictor()
{ }

GMovementPredictor::~GMovementPredictor()
{ }

/** Compute movement and fill in ships' predicted position fields.
    Loosely based upon shipacc.pas::InitMovementPrediction */
void
GMovementPredictor::computeMovement(GUniverse& univ)
{
    init(univ);
    resolveTows(univ);
    while (moveShips(univ)) {
        // nix
    }
    writeBack(univ);
}

/** Get position of a ship. */
GPoint
GMovementPredictor::getShipPosition(int sid) const
{
    if (sid > 0 && sid <= info.size()) {
        return info[sid-1].pos;
    } else {
        return GPoint(0, 0);
    }
}

/** Get cargo of a ship. */
GCost
GMovementPredictor::getShipCargo(int sid) const
{
    if (sid > 0 && sid <= info.size()) {
        return info[sid-1].cargo;
    } else {
        return GCost();
    }
}

/** Initialize movement info.
    Initialize all ship's status and waypoint. */
void
GMovementPredictor::init(const GUniverse& univ)
{
    info.resize(NUM_SHIPS);
    for (int i = univ.ty_any_ships.findNextIndex(0); i != 0; i = univ.ty_any_ships.findNextIndex(i)) {
        const GShip& sh = univ.getShip(i);
        Info& ii = info[i-1];
        ii.status = Normal;
        ii.pos    = sh.isPlayable(GObject::ReadOnly) ? sh.getWaypoint() : sh.getPos();
        copyCargo(ii, sh, el_Tritanium);
        copyCargo(ii, sh, el_Duranium);
        copyCargo(ii, sh, el_Molybdenum);
        copyCargo(ii, sh, el_Supplies);
        copyCargo(ii, sh, el_Money);
    }
}

/** Tow resolution.
    Set all towing/towed ships' status. */
void
GMovementPredictor::resolveTows(const GUniverse& univ)
{
    // Assume any tow succeeds. Anyway, be careful not to make tow groups
    // with more than two ships.
    for (int i = univ.ty_played_ships.findNextIndex(0); i != 0; i = univ.ty_played_ships.findNextIndex(i)) {
        const GShip& sh = univ.getShip(i);

        // Compiler bug: gcc-3.3 generates an infinite loop if tow_id is replaced by sh.getTowId() below.
        int tow_id = sh.getTowId();
        if (sh.getMission() == GMission::msn_Tow
            && univ.ty_any_ships.isValidIndex(tow_id)
            && info[i-1].status == Normal
            && info[tow_id-1].status == Normal)
        {
            // Ship is trying to tow a valid ship,
            // and neither already has a different role in towing.
            info[i-1].status = Towing;
            info[tow_id-1].status = Towed;
        }
    }
}

/** Compute actual movement.
    \retval true There was progress; call again
    \retval false No progress; stop loop */
bool
GMovementPredictor::moveShips(const GUniverse& univ)
{
    bool moved = false;             // true if we moved a ship
    int unresolved_sid = 0;         // ID of a ship with unresolved intercept

    // Try moving all ships.
    // - not ours: just mark it moved.
    // - ours, Normal, not intercepting: move and mark Moved
    // - ours, Towing: move, move towee, mark both Moved
    // - ours, Normal, intercepting, towee Moved: move, mark Moved
    // - ours, Normal, intercepting, towee not Moved: wait for next iteration.
    //   For a normal, non-cyclic intercept, the next iteration will ultimately move it.
    //   For a cyclic intercept, we need special handling; see below.
    for (int sid = univ.ty_any_ships.findNextIndex(0); sid != 0; sid = univ.ty_any_ships.findNextIndex(sid)) {
        const GShip& sh = univ.getShip(sid);
        Info& i = info[sid-1];
        Status ost = i.status;
        if (sh.isPlayable(GObject::ReadOnly) && (ost == Normal || ost == Towing)) {
            // We could possibly move this ship. Intercept?
            if (isValidIntercept(sh, info)) {
                const Info& target = info[sh.getInterceptId()-1];
                if (target.status != Moved) {
                    // This is an unresolved intercept. Save it for later.
                    if (unresolved_sid == 0) {
                        unresolved_sid = sid;
                    }
                    continue;
                }
                i.pos = target.pos;
            }

            // Work on a copy of the ship
            GShipTurnPredictor pred(univ, sid);
            pred.setWaypoint(i.pos);
            pred.computeTurn();
            i.pos = pred.getPosition();
            i.status = Moved;
            copyCargo(i, pred);
            if (ost == Towing) {
                // Move towee
                Info toweeInfo = info[sh.getTowId()-1];
                toweeInfo.pos = i.pos;
                toweeInfo.status = Moved;

                // Compute towee's turn.
                // Normally, we'd have to use the three-argument GShipTurnPredictor
                // to compute both turns at once.
                // However, since we're not interested in fuel usage,
                // we can also compute the towee separately.
                const GShip& towee = univ.getShip(sh.getTowId());
                if (towee.isPlayable(GObject::ReadOnly)) {
                    GShipTurnPredictor toweePred(univ, sh.getTowId());
                    toweePred.setWaypoint(toweePred.getPosition());
                    toweePred.setSpeed(0);
                    toweePred.computeTurn();
                    copyCargo(toweeInfo, toweePred);
                }
            }
        }

        // Mark ship done if possible.
        // If it is being towed, we cannot mark it here;
        // it'll be marked by the above code during a later iteration.
        // Make sure to not mark it done too early, otherwise intercepting towed ships won't work.
        if (i.status != Towed && i.status != Moved) {
            i.status = Moved;
            moved = true;
        }
    }

    // Resolve cyclic intercepts.
    // If we did not move a ship in the above loop, but found one that is intercepting,
    // this means the ship is the entry into an intercept loop
    // (but possibly not part of the actual loop).
    if (!moved && unresolved_sid != 0) {
        // Find all ships that are part of the loop
        int sid = unresolved_sid;
        while (info[sid-1].status == Normal) {
            if (!isValidIntercept(univ.getShip(sid), info)) {
                // This is an error and cannot (should not) happen.
                // If it happens anyway, stop to avoid infinite loop.
                return false;
            }
            info[sid-1].status = ResolvingLoop;
            sid = univ.getShip(sid).getInterceptId();
        }

        // Now, %sid points to a ship which is part of the loop.
        // Go around once more, collecting the coordinates.
        const int loop_start = sid;
        int32_t sum_x = 0;
        int32_t sum_y = 0;
        int num_ships = 0;
        do {
            const GShip& sh = univ.getShip(sid);
            ASSERT(isValidIntercept(sh, info));
            sum_x += sh.getPos().x;
            sum_y += sh.getPos().y;
            ++num_ships;
            sid = sh.getInterceptId();
        } while (sid != loop_start);

        // Resolve the PHost way, very simple: move everyone to the geometric center.
        // We do not care for wrap for now.
        sum_x /= num_ships;
        sum_y /= num_ships;
        do {
            GShipTurnPredictor pred(univ, sid);
            pred.setWaypoint(GPoint(sum_x, sum_y));
            pred.computeTurn();
            info[sid-1].pos = pred.getPosition();
            info[sid-1].status = Moved;
            copyCargo(info[sid-1], pred);
            moved = true;
            sid = univ.getShip(sid).getInterceptId();
        } while (sid != loop_start);

        // Now, clear all other ResolvingLoop ships. These are the entry into the loop;
        // the next iteration will move them because they now have fulfilled preconditions.
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            if (info[i-1].status == ResolvingLoop) {
                info[i-1].status = Normal;
            }
        }
    }

    return moved;
}

/** Write back prediction into main universe. */
void
GMovementPredictor::writeBack(GUniverse& univ)
{
    for (int sid = univ.ty_any_ships.findNextIndex(0); sid != 0; sid = univ.ty_any_ships.findNextIndex(sid)) {
        univ.getShip(sid).setPredictedPos(info[sid-1].pos);
    }
}


/** Check valid intercept.
    The intercept must be in a status that allows us to resolve it,
    i.e. it must not target a nonexisting ship. That is:
    - valid Id
    - not referring to a nonexistant ship
    - not referring to itself (this condition is actually redundant with the current implementation) */
bool
GMovementPredictor::isValidIntercept(const GShip& sh, const InfoVector& info)
{
    int msn = sh.getMission();
    int i = sh.getInterceptId();
    return msn == GMission::msn_Intercept
        && i > 0
        && i <= NUM_SHIPS
        && info[i-1].status != NonExisting
        && i != sh.getId();
}

/** Set cargo amount from ship. */
void
GMovementPredictor::copyCargo(Info& info, const GShip& sh, GCargoType el)
{
    if (sh.isPlayable(GObject::ReadOnly)) {
        info.cargo.set(el, sh.getCargo(el));
    } else {
        info.cargo.set(el, 0);
    }
}

void
GMovementPredictor::copyCargo(Info& info, const GShipTurnPredictor& pred)
{
    info.cargo.set(el_Tritanium,  pred.getCargo(el_Tritanium));
    info.cargo.set(el_Duranium,   pred.getCargo(el_Duranium));
    info.cargo.set(el_Molybdenum, pred.getCargo(el_Molybdenum));
    info.cargo.set(el_Supplies,   pred.getCargo(el_Supplies));
    info.cargo.set(el_Money,      pred.getCargo(el_Money));
}
