/**
  *  \file int/predict.cc
  *  \brief Interpreter: Auto Task Predictions
  *
  *  This interprets a small subset of CCScript, namely just the
  *  "command arg,arg,arg" syntax for simple, literal arguments.
  *  This is what the GUI produces.
  *
  *  Technically, a much larger subset can be used (e.g. variables). 
  *  Also, users can write their own auto task commands which we don't
  *  know anything about.
  *
  *  Note that PCC2 refuses structural commands ('If', 'For') in auto
  *  tasks, cf. IntAutoTaskEditor::isValidCommand(), so we don't have
  *  to deal with them.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/predict.h"
#include "int/lex.h"
#include "int/data.h"
#include "int/autotask.h"
#include "int/args.h"
#include "int/interror.h"

/** Read a single argument. Returns true on success, returns false or
    throws IntError on error.

    argument ::= string
               | bool
               | ["+"|"-"] number

    \param tok [in] Token stream
    \param args [in/out] Arguments collected here  */
static bool
readOneArg(IntTokenizer& tok, IntDataSegment& args)
{
    if (tok.getCurrentToken() == tok.tString) {
        // String
        args.pushNew(new IntStringValue(tok.getCurrentString()));
        tok.readNext();
        return true;
    } else if (tok.getCurrentToken() == tok.tBoolean) {
        // Boolean
        args.pushNew(makeBoolValue(tok.getCurrentInteger()));
        tok.readNext();
        return true;
    } else {
        // Must be number. Let's accept signed numbers, just in case.
        int sign = +1;
        if (tok.checkAdvance(tok.tMinus)) {
            sign = -1;
        } else if (tok.checkAdvance(tok.tPlus)) {
            // nix
        } else {
            // auch nix
        }

        if (tok.getCurrentToken() == tok.tInteger) {
            // integer
            args.pushNew(makeIntValue(sign * tok.getCurrentInteger()));
            tok.readNext();
            return true;
        } else if (tok.getCurrentToken() == tok.tFloat) {
            // float
            args.pushNew(makeFloatValue(sign * tok.getCurrentFloat()));
            tok.readNext();
            return true;
        } else {
            return false;
        }
    }
}

/** Read list of arguments. Returns true on success, returns false or
    throws IntError on error.
    \param tok [in] Token stream
    \param args [out] Arguments */
static bool
readArgs(IntTokenizer& tok, IntDataSegment& args)
{
    if (tok.getCurrentToken() == tok.tEnd) {
        // No args: valid
        return true;
    }

    while (1) {
        // Read one
        if (!readOneArg(tok, args)) {
            return false;
        }

        // Check delimiter
        if (tok.checkAdvance(tok.tComma)) {
            // ok
        } else if (tok.getCurrentToken() == tok.tEnd) {
            // ok
            return true;
        } else {
            return false;
        }
    }
}

/************************** IntAutoTaskPredictor *************************/

IntAutoTaskPredictor::IntAutoTaskPredictor()
{ }

IntAutoTaskPredictor::~IntAutoTaskPredictor()
{ }

/** Predict auto task. Starts interpreting at the current program
    counter, and ends at endPC, if specified.
    - interprets only regular commands with simple arguments;
      no structure commands, no computations
    - performs at most one RESTART loop */
void
IntAutoTaskPredictor::predictTask(const IntAutoTaskEditor& editor, uint32_t endPC)
{
    uint32_t startPC = editor.getPC();
    uint32_t pc = startPC;
    bool looped = false;

    /* Loop until we reach the configured end, or, when we have
       looped, the starting position */
    while (pc < endPC && pc < editor.getNumInstructions() && (!looped || pc < startPC)) {
        try {
            // Parse one line
            IntTokenizer tok(editor[pc]);
            ++pc;

            // Extract command
            if (tok.getCurrentToken() == tok.tEnd) {
                continue;
            }
            if (tok.getCurrentToken() != tok.tIdentifier) {
                break;
            }
            string_t command = tok.getCurrentString();
            tok.readNext();

            // Build arguments
            IntDataSegment args;
            if (!readArgs(tok, args)) {
                break;
            }

            // Call predictor
            if (command == "RESTART") {
                looped = true;
                pc = 0;
            } else {
                IntArgBlock block(args, 0, args.size());
                if (!predictInstruction(command, block)) {
                    break;
                }
            }
        }
        catch (IntError&) {
            break;
        }
    }
}

/** Predict single statement. If the specified program counter points to a
    valid instruction, calls predictInstruction for it, otherwise does nothing.
    Use this to parse single instructions.
    \param editor Editor
    \param pc Instruction to interpret */
void
IntAutoTaskPredictor::predictStatement(const IntAutoTaskEditor& editor, uint32_t pc)
{
    if (pc < editor.getNumInstructions()) {
        try {
            // Parse one line
            IntTokenizer tok(editor[pc]);

            if (tok.getCurrentToken() != tok.tIdentifier) {
                return;
            }
            string_t command = tok.getCurrentString();
            tok.readNext();

            // Build arguments and call predictor
            IntDataSegment args;
            if (readArgs(tok, args) && command != "RESTART") {
                IntArgBlock block(args, 0, args.size());
                predictInstruction(command, block);
            }
        }
        catch (IntError&) { }
    }
}
